{-# LANGUAGE GADTs, ScopedTypeVariables, FlexibleContexts, RankNTypes, TypeFamilies #-}
{-# OPTIONS_GHC -Wall -fno-warn-name-shadowing #-}
module UI.UI where

import Data.IORef
import Graphics.UI.Gtk hiding (Language, Cursor)
import Graphics.UI.Gtk.Glade
import Data.List
import Control.Monad
import Data.Cursor.CLASE.Language
import Data.Cursor.CLASE.Persistence
import Prelude hiding (log)
import Data.Map (Map)
import qualified Data.Map as Map

class (Language l, 
       Persistable l) => UI l where
  
  render :: Cursor l x a -> String

  renderContextInfo :: Cursor l x a -> String
  
data CursorHolder l where
  CH :: (UI l) => Cursor l a a -> Map Int (ExistsR l (Route l a)) -> CursorHolder l

initUI :: (UI l) => Either (Cursor l l l) FilePath -> IO ()
initUI cursorOrfile = do
  initGUI

  Just xml <- xmlNew "UI/language.glade"

  cursorTextView <- xmlGetWidget xml castToTextView "cursor"
  cursorBuffer <- textViewGetBuffer cursorTextView

  inContextTextView <- xmlGetWidget xml castToTextView "inContext"
  inContextBuffer <- textViewGetBuffer inContextTextView

  saveButton <- xmlGetWidget xml castToButton "saveButton"
  savePath <- xmlGetWidget xml castToEntry "savePath"

  savePath `entrySetText` (either (const "/tmp/foo") id cursorOrfile)

  (RestoredCursor cursor) <- getCursor cursorOrfile
  let guiState = GS {
    cursorBuffer = cursorBuffer,
    inContextBuffer = inContextBuffer,
    cursorH = CH (resetLog cursor) Map.empty
  }

  ref <- newIORef guiState

  saveButton `onClicked` (saveToPath savePath ref)
  cursorTextView `onKeyPress` (cursorKeyPress ref)
  refreshAll ref

  window <- xmlGetWidget xml castToWindow "mainWindow"
  onDestroy window mainQuit
  widgetShowAll window
  
  mainGUI
  

saveToPath :: Entry -> IORef (GuiState l) -> IO ()
saveToPath entry ref  = do
  GS { cursorH = CH cursor _ } <- readIORef ref
  fp <- entryGetText entry
  writeFile fp (showCursor cursor)

getCursor :: (Persistable l) => Either (Cursor l l l) FilePath -> IO (RestoredCursor l)
getCursor (Left c) = return $ RestoredCursor c
getCursor (Right p) = do
  str <- readFile p
  let Just rc = parseCursor str
  return rc

cursorKeyPress :: forall l . (Language l) => IORef (GuiState l) -> Event -> IO Bool
cursorKeyPress ref (Key { eventKeyChar = Just char })
  | char == 'h' = moveCursor [gml, gmu]
  | char == 'j' = moveCursor [gmd, gmr]
  | char == 'k' = moveCursor [gmu, gml]
  | char == 'l' = moveCursor [gmr, gmd]
  | char `elem` loadBookmark = do
      let (Just bid) = findIndex (== char) loadBookmark
      gs@GS { cursorH = CH theCursor @ Cursor {} bmap } <- readIORef ref
      maybe (return True)
            (\(ExistsR bmRoute) -> do
              let (Just cursor') = followRoute theCursor bmRoute
              let bmap' = Map.map (\(ExistsR r) -> ExistsR $ (log cursor') `appendRoute` r) bmap
              writeIORef ref (gs { cursorH = CH (resetLog cursor') bmap' })
              refreshAll ref
              return True
            )
            (Map.lookup bid bmap)

  | char `elem` saveBookmark = do
      let (Just bid) = findIndex (== char) saveBookmark
      gs@GS { cursorH = CH theCursor @ Cursor {} bmap } <- readIORef ref
      let bmap' = Map.insert bid (ExistsR (Route Stop Stop)) bmap
      writeIORef ref (gs { cursorH = CH theCursor bmap' })
      return True
  where
    loadBookmark = "123456789"
    saveBookmark = "!\"£$%^&*("
    gmd = fmap (\(CWM c _) -> ExistsR c) . genericMoveDown
    gmu = fmap (\(CWM c _) -> ExistsR c) . genericMoveUp
    gml = genericMoveLeft
    gmr = genericMoveRight

    moveCursor :: (forall x a . [Cursor l x a -> Maybe (ExistsR l (Cursor l x))]) -> IO Bool
    moveCursor movs = do
      gs@GS { cursorH = CH theCursor @ Cursor {} bmap} <- readIORef ref

      maybe (return True) 
            (\(ExistsR cursor') -> do
              let bmap' = Map.map (\(ExistsR r) -> ExistsR $ (log cursor') `appendRoute` r) bmap
              writeIORef ref (gs { cursorH = CH (resetLog cursor') bmap' })
              refreshAll ref
              return True
            )
            (msum $ map (\f -> f theCursor) movs)
cursorKeyPress _ _ = return False

refreshAll :: IORef (GuiState l) -> IO ()
refreshAll ref = do
  gs@GS { cursorH = CH cursor@Cursor {} _ } <- readIORef ref

  let cursorText = render cursor
  let contextInfo = renderContextInfo cursor

  (cursorBuffer gs) `textBufferSetText` cursorText
  (inContextBuffer gs) `textBufferSetText` contextInfo

data GuiState l = GS {
  cursorBuffer :: TextBuffer,
  inContextBuffer :: TextBuffer,
  cursorH :: CursorHolder l
}
